#include <Arduino.h>
#include "MIDIUSB.h"
#include <Multiplexer4067.h>  // Multiplexer CD4067 library >> https://github.com/sumotoy/Multiplexer4067
#include <Thread.h>           // Threads library >> https://github.com/ivanseidel/ArduinoThread
#include <ThreadController.h> // Same as above
#include <Wire.h>
#include "oled.h"
#include "PotDisplay.h"

// uncomment for Serial debug output
// #define DEBUG 1

// initializes display and rendering routine
OLED oled(Wire, 0x3C, NO_RESET_PIN, 128, 64, false); // true=SH1106 chip, false=SSD1306 chip
PotDisplay potDisplay(oled);

#define N_MUX 1 // number of multiplexers
//* Define s0, s1, s2, s3, and x pins
#define s0 9
#define s1 8
#define s2 7
#define s3 6
#define x1 A0 // analog pin of the first mux

// initializes the multiplexer
Multiplexer4067 mux[N_MUX] = {
    Multiplexer4067(s0, s1, s2, s3, x1)};

// forward declarations of thread functions
void potentiometers();
void renderFrame();

// POTENTIOMETERS
const int N_POTS = 0 + 16;                       // total numbers of pots (slide & rotary). Number of pots in the Arduino + number of pots on multiplexer 1 + number of pots on multiplexer 2...
const int N_POTS_ARDUINO = 0;                   // number of pots connected straight to the Arduino
const int POT_ARDUINO_PIN[N_POTS_ARDUINO] = {}; // pins of each pot connected straight to the Arduino

const int N_POTS_PER_MUX[N_MUX] = {16}; // number of pots in each multiplexer (in order)
const int POT_MUX_PIN[N_MUX][16] = {
    // pins of each pot of each mux in the order you want them to be
    {0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15} // pins of the first mux
};

byte POT_CC_N[N_POTS] = {4, 5, 6, 21, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 20};

// display strings for each pot / CC mapping
// \n to devide the two lines
// each line holds about 11/12 characters
const char string_1[] PROGMEM = "Expression";
const char string_2[] PROGMEM = "Vibrato\nDepth";
const char string_3[] PROGMEM = "Vibrato\nRate";
const char string_4[] PROGMEM = "Portamento";
const char string_5[] PROGMEM = "Harmonic\nStructure";
const char string_6[] PROGMEM = "Sub\nHarmonics";
const char string_7[] PROGMEM = "Growl";
const char string_8[] PROGMEM = "Flutter\nTongue";
const char string_9[] PROGMEM = "Formant";
const char string_10[] PROGMEM = "Breath\nNoise";
const char string_11[] PROGMEM = "Key\nNoise";
const char string_12[] PROGMEM = "Dynamic\nPitch";
const char string_13[] PROGMEM = "Pan Pot";
const char string_14[] PROGMEM = "Reverb\nMix";
const char string_15[] PROGMEM = "CC 17";
const char string_16[] PROGMEM = "Volume";

const char *const POT_CC_STRINGS[] PROGMEM =
    {
        string_1, string_2, string_3, string_4, string_5, string_6, string_7, string_8,
        string_9, string_10, string_11, string_12, string_13, string_14, string_15, string_16};

int potCState[N_POTS] = {0}; // Current state of the pot
int potPState[N_POTS] = {0}; // Previous state of the pot
int potVar = 0;              // Difference between the current and previous state of the pot

int potMidiCState[N_POTS] = {0}; // Current state of the midi value
int potMidiPState[N_POTS] = {0}; // Previous state of the midi value

int MIDI_CH = 1;                   // MIDI channel to be used
byte LED_PIN = 3;                  // System up indicator
const int TIMEOUT = 200;           // Amount of time the potentiometer will be read after it exceeds the varThreshold
const int varThreshold = 10;       // Threshold for the potentiometer signal variation
boolean potMoving = true;          // If the potentiometer is moving
unsigned long PTime[N_POTS] = {0}; // Previously stored time
unsigned long timer[N_POTS] = {0}; // Stores the time that has elapsed since the timer was reset
bool firstRun = true;              // Prevent displaying values on startUp

// THREADS
// This libs create a "fake" thread. This means you can make something happen every x milisseconds
// We can use that to read something in an interval, instead of reading it every single loop
// In this case we'll read the potentiometers in a thread, making the reading of the buttons faster
ThreadController cpu;        // thread master, where the other threads will be added
Thread threadPotentiometers; // thread to control the pots
Thread threadRendering;      // thread to render display frames

void setup()
{
  // Baud Rate
  Serial.begin(115200); // 31250 for MIDI class compliant | 115200 for Hairless MIDI
#ifdef DEBUG
  while (!Serial)
  {
    ; // wait for serial port to connect
  }
  Serial.println("Serial ready");
#endif

  for (int i = 0; i < N_MUX; i++) // initialize multiplexers
  {
    mux[i].begin();
  }

  pinMode(x1, INPUT_PULLUP); // set each X pin as input_pullup (avoid floating behavior)

  // initialize display
  Wire.begin();
  Wire.setClock(400000);
  oled.init();

  // initialize threads
  threadPotentiometers.setInterval(10);       // every how many milliseconds
  threadPotentiometers.onRun(potentiometers); // the function that will be added to the thread

  threadRendering.setInterval(1000 / 50); // 50 frames per second
  threadRendering.onRun(renderFrame);

  cpu.add(&threadPotentiometers); // add every thread here
  cpu.add(&threadRendering);      // add every thread here

  pinMode(LED_PIN, OUTPUT);
  analogWrite(LED_PIN, 100);
}

void loop()
{
  cpu.run(); // for threads
}

// POTENTIOMETERS THREAD
void potentiometers()
{
  // reads the pins from arduino
  for (int i = 0; i < N_POTS_ARDUINO; i++)
  {
    potCState[i] = analogRead(POT_ARDUINO_PIN[i]);
  }

  // It will happen if using a mux
  int nPotsPerMuxSum = N_POTS_ARDUINO; // offsets the buttonCState at every mux reading

  // reads the pins from every mux
  for (int j = 0; j < N_MUX; j++)
  {
    for (int i = 0; i < N_POTS_PER_MUX[j]; i++)
    {
      potCState[i + nPotsPerMuxSum] = mux[j].readChannel(POT_MUX_PIN[j][i]);
    }
    nPotsPerMuxSum += N_POTS_PER_MUX[j];
  }

  for (int i = 0; i < N_POTS; i++)
  { // Loops through all the potentiometers

    potMidiCState[i] = constrain(map(potCState[i], 10, 1013, 0, 127), 0, 127); // Maps the reading of the potCState to a value usable in midi
    potVar = abs(potCState[i] - potPState[i]);                                 // Calculates the absolute value between the difference between the current and previous state of the pot

    if (potVar >= varThreshold)
    {                      // Opens the gate if the potentiometer variation is greater than the threshold
      PTime[i] = millis(); // Stores the previous time
    }

    timer[i] = millis() - PTime[i]; // Resets the timer 11000 - 11000 = 0ms

    if (timer[i] < TIMEOUT)
    { // If the timer is less than the maximum allowed time it means that the potentiometer is still moving
      potMoving = true;
    }
    else
    {
      potMoving = false;
    }
    if (potMoving == true)
    { // If the potentiometer is still moving, send the change control
      if (potMidiPState[i] != potMidiCState[i])
      {

        // Sends the MIDI CC
        midiEventPacket_t event = {0x0B, 0xB0 | (uint8_t)MIDI_CH, (uint8_t)POT_CC_N[i], (uint8_t)potMidiCState[i]};
        MidiUSB.sendMIDI(event);
        MidiUSB.flush();

        // output value on display
        if (!firstRun)
        {
          potDisplay.render(POT_CC_N[i], potMidiCState[i], &POT_CC_STRINGS[i]);
        }

        // Stores the current reading of the potentiometer to compare with the next
        potPState[i] = potCState[i];
        potMidiPState[i] = potMidiCState[i];

#ifdef DEBUG
        char output[60];
        sprintf(output, "pot %d analog %u - midi %u - moving %d firstRun %d", i, potCState[i], potMidiCState[i], potMoving, firstRun);
        Serial.println(output);
#endif
      }
    }
  }
  firstRun = false;
}

// RENDERING FRAMES THREAD
void renderFrame()
{
  potDisplay.renderFrame();
}